/**********************************************************************/
/*                                                                    */
/*  File:          xmixer.c                                           */
/*                                                                    */
/**********************************************************************/

#include "config.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <linux/soundcard.h>

#include "McApp.h"
#include "McGadget.h"
#include "McSlider.h"
#include "McText.h"
#include "McBitmap.h"
#include "icons.h"
#include "xmixer.h"

#define BIG_MIXER_CONTROLS    { 1,1,1,1,1,1,1,1,1,1,1,1 }
#define SMALL_MIXER_CONTROLS  { 1,1,1,1,1,1,0,0,1,0,0,0 }
#define STEREO                { 1,0,0,1,1,0,1,1,1,1,1,1 }

#define SLIDER_DISTANCE        45
#define SLIDER_OFFSET_MONO     6
#define SLIDER_SIZE_STEREO     25
#define SLIDER_SIZE_MONO       14

/* Private routines
 * ================
 */
static void slider_proc(McGadget *gadget);
static void button_proc(McGadget *gadget);
static void record_proc(McGadget *gadget);
static void exit_proc(McGadget *gadget);
static void alarm_proc();
static void main_loop();
static int get_volume(McGadget *gadget);
static int set_volume(McGadget *gadget);
static int set_get_source(int nr, int onoff);
static int set_flag(McGadget *gadget);
static int get_flag(McGadget *gadget);
static void AddSwitch(char *title, McBitmap *bm1, McBitmap *bm2, int request);
static McGadget *AddSlider(int x,int y,int request,int stereo);
static McGadget *AddIcon(int x,int y, int w, int h, int request);
static McGadget *AddSelector(int x, int y, int request);
static int ScanArg(int argc, char *argv[], int i);
static void deactivate(McGadget *gadget);

/* Private variables
 * =================
 */


/* Public routines
 * ===============
 */


/* Public variables
 * ================
 */
char		*icon_name;

char		window_title[128], *window_name = window_title;
char		*xmixer_version;
char            default_mixer[]="/dev/mixer";
char            *mixer_name=default_mixer;
char            *labels[] = SOUND_DEVICE_LABELS;
McGadget	*Gad[NUM_GADGETS];
McApp		*app;

int             mixer;
int		GadCnt = 0;
int		WindowVisible = 0;
int             small_mix_control[] = SMALL_MIXER_CONTROLS;
int             big_mix_control[]   = BIG_MIXER_CONTROLS;
int             *control;
int             stereo[] = STEREO;
int             WindowWidth, WindowHeight;
int		mode = 0, messagenull = 0;
int		source = 0; /* contains a bit for each supported recording
			       source */
#ifdef POLL_MIXER
int		poll_mixer = 1;
#else
int		poll_mixer = 0;
#endif

#define MAX_SLIDERS 12

/**************************************************************************
 * main(int ac, char **argv)
 *
 */
main(int argc, char **argv) {
  int	i,j;
  char	*display_name = NULL;
  Pixmap  icon_pixmap;
  Pixmap iconmask_pixmap;
  struct sigaction alarm_sigaction= { alarm_proc, 0, 0, 0 };
  McGadget *gad;

  xmixer_version = "XMixer V1.1";
  myname = argv[0];

  /*
   * Build connection with X server
   */
  app = McCreateApp(McConnectToServer(&argc, &argv, ScanArg));
  if (argc>1) {
    fprintf(stderr,"%s: Warning: Garbage on command line.\n",myname);
  }

  /*
   * Open the mixer device
   */
  if ((mixer = open(mixer_name,O_RDWR))<0) {
    fprintf(stderr,"%s: Can't open %s: ",argv[0],mixer_name);
    perror("");
    cleanup(1);
  }

  /*
   * Determine the available recording sources
   */

  if (ioctl(mixer, SOUND_MIXER_READ_RECMASK, &source)<0) {
    fprintf(stderr,
	    "%s: Warning: Can't get a list of available recording sources.\n",
	    myname);
    source=0;
  }

  switch(mode) {
  case 1:
    control = small_mix_control;
    icon_name = "XMixer";
    break;
  default:
    control = big_mix_control;
    icon_name = "XMixer";
    break;
  }

#ifndef NO_ALTPCMDEVICE
#ifdef ALTPCMDEVICE
  if (control[SOUND_MIXER_ALTPCM]) { /* Check if the second device exists */
    int x;
    if ((x=open(ALTPCMDEVICE, O_RDONLY))<0) {
      control[SOUND_MIXER_ALTPCM]=0;
    } else
      close(x);
  }
#endif
#else /*NO_ALTPCMDEVICE*/
  control[SOUND_MIXER_ALTPCM]=0;
#endif /*NO_ALTPCMDEVICE*/

  strcpy(window_title, icon_name);
  WindowWidth=12;
  for (i=0;i<MAX_SLIDERS;i++) {
    if (control[i])
      WindowWidth+=SLIDER_DISTANCE;
  }
  if (WindowWidth<278) WindowWidth=278;
  WindowHeight = 230;

  /*
   * Create the window
   */
  McCreateAppWindow(app, WindowWidth, WindowHeight);

  /*
   * Create all bitmaps
   */

  for (i=NUM_BITMAPS; --i>=0;)
    pmap[i].pixmap = XCreateBitmapFromData(app->display,
					   app->window, pmap[i].bits, 
					   pmap[i].width, pmap[i].height);

  /*
   * Set WM Hints
   */
 {
   XWMHints	 wm_hints;
   XClassHint	 class_hints;
   XTextProperty windowName;
   XTextProperty iconName;
   XSizeHints	 sizehints;

   if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
     fprintf(stderr, "%s: structure allocation for windowName failed.\n",
	     argv[0]);
     cleanup(1);
   }
   if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
     fprintf(stderr, "%s: structure allocation for iconName failed.\n",
	     argv[0]);
     cleanup(1);
   }

   wm_hints.initial_state = NormalState;
   wm_hints.input = True;
   wm_hints.icon_pixmap = pmap[PIX_ICON].pixmap;
   wm_hints.icon_mask = pmap[PIX_ICONMASK].pixmap;
   wm_hints.flags = StateHint | IconPixmapHint | IconMaskHint | InputHint;
   class_hints.res_name = argv[0];
   class_hints.res_class = "XMixer";
   sizehints.flags = PSize | PMinSize | PMaxSize | PResizeInc;
   sizehints.min_width = 0;
   sizehints.width = sizehints.max_width = WindowWidth;
   sizehints.height= sizehints.max_height= WindowHeight;
   sizehints.min_height= 0;
   sizehints.width_inc = sizehints.height_inc = 1;
   
   XSetWMProperties(app->display, app->window, &windowName, &iconName,
		    argv, argc, &sizehints, &wm_hints, &class_hints);
 }

  /*
   * Select the input events we will be wanting to deal with
   */
  XSelectInput(app->display, app->window, 
	       ExposureMask |
	       KeyPressMask |
	       ButtonPressMask |
	       ButtonReleaseMask |
	       ButtonMotionMask |
	       StructureNotifyMask);

  /*
   * Display the window
   */
  XMapWindow(app->display, app->window);
  XFlush(app->display);

  Gad[0] = McCreateGadget(app, GAD_HBOX, BOOLGADGET,
			  10, WindowHeight - 20, 200, 16);
  Gad[0]->normalLabel = McCreateText(app, "", app->gc[GC_NORMAL],
				     app->normalFont, 4, 3);

  Gad[1] = McCreateGadget(app, GAD_H3D | GAD_3D | GAD_ACTIVE, BOOLGADGET,
			  10, 11, 39, 20);
  Gad[1]->normalLabel = McCreateText(app, "Exit", app->gc[GC_NORMAL],
				     app->normalFont, 8, 4);
  Gad[1]->callbackUp = exit_proc;

  GadCnt=2;
  AddSwitch("Loudness",NULL,NULL,SOUND_MIXER_LOUD);
  if (get_flag(Gad[GadCnt-1])<0) deactivate(Gad[GadCnt-1]);
  AddSwitch(NULL,&pmap[PIX_NOWIDE],&pmap[PIX_WIDE],SOUND_MIXER_ENHANCE);
  if (get_flag(Gad[GadCnt-1])<0) deactivate(Gad[GadCnt-1]);
  AddSwitch(NULL,&pmap[PIX_NOMUTE],&pmap[PIX_MUTE],SOUND_MIXER_MUTE);
  if (get_flag(Gad[GadCnt-1])<0) deactivate(Gad[GadCnt-1]);

  for (i=0,j=11;i<MAX_SLIDERS;i++) {
    if (control[i]) {
      AddIcon(j, 42, 39, 165, i);
      j+=SLIDER_DISTANCE;
    } else {
      Gad[GadCnt]=NULL;
      GadCnt++;
    }
  }

  if (Gad[ICO_IMIX] && Gad[ICO_ALTPCM]) { /* Swap them on the screen */
    int n;
    n=Gad[ICO_IMIX]->x;
    Gad[ICO_IMIX]->x = Gad[ICO_ALTPCM]->x;
    Gad[ICO_ALTPCM]->x=n;
    n=Gad[ICO_IMIX]->y;
    Gad[ICO_IMIX]->y = Gad[ICO_ALTPCM]->y;
    Gad[ICO_ALTPCM]->y=n;
  }

  for (i=0;i<MAX_SLIDERS;i++) {
    if (control[i]) {
      gad = Gad[i + ICO_VOLUME];
      AddSlider(gad->x + 7, gad->y + 35, i, stereo[i]);
      get_volume(Gad[GadCnt-1]);
    } else {
      Gad[GadCnt]=NULL;
      GadCnt++;
    }
  }

  for (i=0;i<MAX_SLIDERS;i++) {
    if (control[i] && (source&(1<<i))) {
      gad = Gad[i + ICO_VOLUME];
      AddSelector(gad->x + 7, gad->y + 143, i);
    } else {
      Gad[GadCnt]=NULL;
      GadCnt++;
    }
  }

  /* Read the current source */
  set_get_source(-1, 0);

  /*
   * Restore the message after a delay
   */
  sigaction(SIGALRM,&alarm_sigaction,0);

  /*
   * Now do the main loop
   */
  main_loop();
}

static int ScanArg(int argc, char *argv[], int i) {
  if (!strcmp(argv[i], "-0")) {
    mode = 0;
    return 1;
  }
  if (!strcmp(argv[i], "-1")) {
    mode = 1;
    return 1;
  }
  if (!strcmp(argv[i], "-p")) {
    poll_mixer = 1;
    return 1;
  }
  if (!strcmp(argv[i], "-np")) {
    poll_mixer = 0;
    return 1;
  }
  if (!strcmp(argv[i], "-v")) {
    printf("%s\n", xmixer_version);
    cleanup(0);
  }
  return 0;
}

/**************************************************************************
 * PRIVATE void slider_proc(Slider *slider)
 */

#define SLIDE ((McSlider *)(gadget->specialInfo))
static void slider_proc(McGadget *gadget) {
  set_volume(gadget);
  get_volume(gadget);
  McGadgetUpdate(app, gadget);
}

/**************************************************************************
 * static void exit_proc(McGadget *gadget)
 *
 */
static void exit_proc(McGadget *gadget) {
  cleanup(0);
}

cleanup(int r) {
  McFreeApp(app);
  if (mixer>=0)
    close(mixer);
  exit(r);
}

/**************************************************************************
 * static void button_proc(Button *button)
 *
 */
static void button_proc(McGadget *gadget) {
  if (set_flag(gadget)<0)
    deactivate(gadget);
  else
    if (get_flag(gadget)<0)
      deactivate(gadget);
  McGadgetRedraw(app, gadget);
}

/**************************************************************************
 * static void button_proc(Button *button)
 *
 */
static void record_proc(McGadget *gadget) {
  set_get_source(gadget->id, gadget->flags & GAD_SELECTED);
}

/**************************************************************************
 * static void alarm_proc(..something..)
 *
 */
static void alarm_proc() {
  /* Xlib is not re-entrant */
  messagenull=1;
}

/***************************************************************************/

static void slider_update() {
  int i,f,j;
  f=0;
  for (i=SLD_VOLUME;i<=SLD_RECLEV; i++) {
    if (Gad[i] && (Gad[i]->flags & GAD_ACTIVE)) {
      if ((j=get_volume(Gad[i]))>0) {
	McGadgetRedraw(app, Gad[i]);
	f=1;
      } else {
	if (j<0) {
	  deactivate(Gad[i]);
	  f=1;
	}
      }
    }
  }
  for (i=BTN_LOUD;i<=BTN_MUTE; i++) {
    if (Gad[i] && (Gad[i]->flags & GAD_ACTIVE)) {
      if ((j=get_flag(Gad[i]))>0) {
	McGadgetRedraw(app, Gad[i]);
	f=1;
      } else {
	if (j<0) {
	  deactivate(Gad[i]);
	  f=1;
	}
      }
    }
  }
  if (f) XFlush(app->display);
}

/**************************************************************************/

static void deactivate(McGadget *gadget) {
  gadget->flags&=~GAD_ACTIVE; gadget->flags|=GAD_DISABLE;
}

/**************************************************************************
 * static void main_loop()
 *
 */

void myXNextEvent(Display *display, XEvent *event_return) {
  struct timeval to;
  fd_set rdfds, usefds;
  int dfd;

  if (XPending(display)>0) {
    XNextEvent(display, event_return);
    return;
  }

  FD_ZERO(&rdfds);
  dfd=ConnectionNumber(display);
  FD_SET(dfd,&rdfds);

  while(1) {
    to.tv_sec  = 0;
    to.tv_usec =  MIXER_POLL_RATE;
    usefds=rdfds;
    if (WindowVisible && poll_mixer) {
      select(dfd+1, &usefds, (fd_set *) NULL, (fd_set *) NULL, &to);
    } else {
      select(dfd+1, &usefds, (fd_set *) NULL, (fd_set *) NULL,
	     (struct timeval *) NULL);
    }
    if (messagenull) { messagenull=0; do_message(0); }
    if (XPending(display)>0) {
      XNextEvent(display, event_return);
      return;
    }
    slider_update();
  }
}

static void main_loop() {
  XEvent report, peek;
  int i;

  while(1) {
    myXNextEvent(app->display, &report);
    switch (report.type) {
    case Expose:
      McAppDoEvent(app, &report);
      break;
    case UnmapNotify:
      WindowVisible = 0;
      break;
    case MapNotify:
      slider_update();
      WindowVisible = 1;
      break;
    default:
      McAppDoEvent(app, &report);
      break;
    }
  }
}

/*************************************************************************
 * STATUS STATUS STATUS STATUS STATUS STATUS STATUS STATUS STATUS STATUS */

do_message(char *string) {
  int width, sleep;
	
  if (string) alarm(3); else string="";

  McChangeText(app, Gad[MSG_STATUS]->normalLabel, string,
	       Gad[MSG_STATUS]->normalLabel->gc);
  McGadgetRedraw(app, Gad[MSG_STATUS]);
}

/**************************************************************************/

static McGadget *AddSlider(int x,int y,int request,int stereo) {
  Gad[GadCnt] = McCreateGadget(app,
			       GAD_3D | GAD_H3D | GAD_ACTIVE | GAD_SELECTED,
			       SLIDERGADGET, x+(stereo?0:SLIDER_OFFSET_MONO),
			       y, stereo?SLIDER_SIZE_STEREO:SLIDER_SIZE_MONO,
			       100);
  Gad[GadCnt]->specialInfo = McCreateSlider((stereo?SLIDER_STEREO:SLIDER_MONO)|
					    SLIDER_VERTICAL, 0, 100);

  ((McSlider *)(Gad[GadCnt]->specialInfo))->step = 3;
  Gad[GadCnt]->callbackDown = slider_proc;
  Gad[GadCnt]->id = request;
  GadCnt++;
}

static McGadget *AddIcon(int x,int y, int w, int h, int request) {
  McBitmap *bm;
  bm = &pmap[PIX_VOL + request];
  Gad[GadCnt] = McCreateGadget(app, GAD_3D | GAD_NOFILL, BOOLGADGET,
			       x-1 , y-1,
			       w?w:(bm->width+2), h?h:(bm->height+2));
  Gad[GadCnt]->normalBitmap=bm;
  GadCnt++;
}

static McGadget *AddSelector(int x,int y, int request) {
  Gad[GadCnt] = McCreateGadget(app,GAD_3D | GAD_H3D | GAD_ACTIVE | GAD_TOGGLE,
			       BOOLGADGET, x , y, 25, 16);
  Gad[GadCnt]->id = request;
  Gad[GadCnt]->normalBitmap = &pmap[PIX_MIX_PLAY];
  Gad[GadCnt]->selectBitmap = &pmap[PIX_MIX_REC];
  Gad[GadCnt]->callbackDown = record_proc;
  GadCnt++;
}

static void AddSwitch(char *title, McBitmap *bm1, McBitmap *bm2, int request) {
  Gad[GadCnt] = McCreateGadget(app, GAD_3D | GAD_H3D | GAD_ACTIVE | GAD_TOGGLE,
			       BOOLGADGET,
			       Gad[GadCnt-1]->x + Gad[GadCnt-1]->width + 10,
			       Gad[GadCnt-1]->y, 32, 20);
  Gad[GadCnt]->callbackDown = button_proc;
  if (title) {
    Gad[GadCnt]->normalLabel=McCreateText(app,title,app->gc[GC_NORMAL],
					  app->normalFont, 8, 4);
    Gad[GadCnt]->width = Gad[GadCnt]->normalLabel->width + 16;
  }
  if (bm1) {
    Gad[GadCnt]->normalBitmap=bm1;
    Gad[GadCnt]->width = bm1->width + (bm1->x<<1);
  }
  if (bm2) {
    Gad[GadCnt]->selectBitmap=bm2;
    Gad[GadCnt]->width = bm2->width + (bm2->x<<1);
  }
  Gad[GadCnt]->id = request;
  GadCnt++;
}

/**************************************************************************
 * Device access
 *
 */

static int set_volume(McGadget *gadget) {
  unsigned int n;
  n = ((McSlider *)(gadget->specialInfo))->leftValue +
     (((McSlider *)(gadget->specialInfo))->rightValue<<8);
  if ((mixer<0) || ioctl(mixer,MIXER_WRITE(gadget->id),&n)<0) {
    do_message("Can't set volume.");
    return -1;
  }
  return 0;
}


static int get_volume(McGadget *gadget) {
  unsigned int n;

  if ((mixer<0) || ioctl(mixer,MIXER_READ(gadget->id),&n)<0) {
    do_message("Can't read volume.");
    return -1;
  }
  if ((((McSlider *)(gadget->specialInfo))->leftValue) == (n & 0x7F))
    if ((((McSlider *)(gadget->specialInfo))->rightValue) == ((n>>8) & 0x7F)) {
      return 0;
    }
  ((McSlider *)(gadget->specialInfo))->leftValue = n & 0x7F;
  ((McSlider *)(gadget->specialInfo))->rightValue = (n>>8) & 0x7F;
  return 1;
}

static int set_flag(McGadget *gadget) {
  int n;

  if (gadget->flags&GAD_SELECTED) n=80; else n=0;
  if ((mixer<0) || ioctl(mixer,MIXER_WRITE(gadget->id),&n)<0) {
    do_message("Can't write setting.");
    return -1;
  }
  return 0;
}

static int get_flag(McGadget *gadget) {
  int n;
  if ((mixer<0) || (ioctl(mixer,MIXER_READ(gadget->id),&n)<0)) {
    do_message("Can't read setting.");
    return -1;
  }
  if (n) {
    if (gadget->flags&GAD_SELECTED)
      return 0;
    gadget->flags|=GAD_SELECTED;
  } else {
    if (!(gadget->flags&GAD_SELECTED))
      return 0;
    gadget->flags&=~GAD_SELECTED;
  }
  return 1;
}

static int set_get_source(int nr, int onoff) {
  int n,i,s;

  if (mixer<0) goto bad;

  if (nr>=0) {
    if (ioctl(mixer, SOUND_MIXER_READ_RECSRC, &n)<0) goto bad;
    if (onoff) n |= 1<<nr; else n &= ~(1<<nr);
    if (ioctl(mixer, SOUND_MIXER_WRITE_RECSRC, &n)<0) goto bad;
  }

  if (ioctl(mixer, SOUND_MIXER_READ_RECSRC, &n)<0) goto bad;

  for (i=SEL_VOLUME;i<=SEL_RECLEV;i++) {
    if (n & 1) {
      if (Gad[i] && (!(Gad[i]->flags & GAD_SELECTED))) {
	Gad[i]->flags |= GAD_SELECTED;
	McGadgetRedraw(app, Gad[i]);
      }
    } else {
      if (Gad[i] && (Gad[i]->flags & GAD_SELECTED)) {
	Gad[i]->flags &= ~GAD_SELECTED;
	McGadgetRedraw(app, Gad[i]);
      }
    }
    n=n>>1;
  }
  return 0;

bad:
  do_message("Can't access recording source.");
  for (i=SEL_VOLUME;i<=SEL_RECLEV;i++) {
    if (Gad[i] && (Gad[i]->flags&GAD_ACTIVE)) {
      deactivate(Gad[i]);
      McGadgetRedraw(app, Gad[i]);
    }
  }
  return -1;
}

