/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "X11/Xlib.h"
#include <stdio.h>
#include <stdlib.h>
#include "struct.h"
#include "McApp.h"
#include "McBitmap.h"
#include "McGadget.h"
#include "icons.h"
#include <signal.h>
#include "cd.h"

/*   This file contains code taken from the
     WorkBone CD-Player, Copyright (C) 1993 Thomas McWilliams

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2, or (at your option)
     any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

char *cur_trackname;		/* Take a guess */
int cur_index = 0;		/* Current index mark */
int cur_frame;			/* Current frame number */
int cue = 0, speed = 0;
struct play *playlist = NULL;
struct cdinfo thiscd, *cd = &thiscd;
int cur_track = 1;		/* Current track number, starting at 1 */
int save_track = 1;
char *cur_artist;		/* Name of current CD's artist */
char cur_avoid;			/* Avoid flag */
char cur_contd;			/* Continued flag */
char *cur_cdname;		/* Album name */
int cur_nsections;		/* Number of sections currently defined */
int exit_on_eject = 0;
int cd_stopped_at_end = 0;
int shuffle = 0;
int repeat = 0;
int clk_mode = CLK_TRACK;

int error = 0, enable_eject = 1;
int cur_balance = 10, info_modified;
int cur_track, cur_pos_abs, cur_pos_rel, cur_tracklen, cur_cdlen, cur_cdmode,
  cur_ntracks, cur_lasttrack, cur_firsttrack, cur_listno;
char tmptime[100];
char *tottime;

void alarm_handler() {
  error=0;
}

void cd_init() {
  struct sigaction alarm_sigaction= { alarm_handler, 0, 0, 0 };

  sigaction(SIGALRM,&alarm_sigaction,0);
  thiscd.trk = NULL;
  thiscd.lists = NULL;
}

int set_error(int e) {
  if (!e) return 0;
  error = e;
  cd_stopped_at_end=0;
  alarm(2);
}

/* NeedToPoll becomes TRUE if we have to
 * poll even when the window is iconified
 */
inline void UpdateNeedToPoll() {
  NeedToPoll = (shuffle || repeat) && (cur_cdmode == CDPLAY);
}

/*	0	No CD in drive.
 *	1	CD in drive.
 *	2	CD has just been inserted (TOC has been read)
 *	3	CD has just been removed
 */

void cd_action(int request, int down) {
  int cds, tmppos;

  /********    AUTOMATICS    ************/


  cds=cd_status();
#ifdef DBG_STATUS
  fprintf(stderr, "Status: %d\n",cds);
#endif

  if (cds==4) return; /* do nothing with a mounted cd */

  if (cds==2) save_track = 1;

  if (cur_cdmode==CDPLAY) {
    save_track = cur_track;
    cd_stopped_at_end = 1;
  }

  if (cur_cdmode==CDSTOP && cd_stopped_at_end && repeat) {
    save_track = 1;
    cur_track = 1;
    set_error(play_cd (1, 0, cur_ntracks + 1));
  }

  if (cur_cdmode==CDNULL) cur_track = 1;



  /* If we are not visible, no need to process further */
  if (!WindowVisible) return;



  if ((cds==2) || (cds==1)) {

    if (request==REQ_POLL) {
#ifdef DBG_STATUS
      fprintf(stderr, "poll... cue:%d  mode:%d\n",cue, cur_cdmode);
#endif


      /********    FAST SEARCH    ************/

      if ((((cue>0) && (Gad[BTN_FF]->flags&GAD_SELECTED)) ||
	   ((cue<0) && (Gad[BTN_REWIND]->flags&GAD_SELECTED))) &&
	  (cur_cdmode==CDPLAY)) {
#ifdef DBG_STATUS
	fprintf(stderr, "pos: %d\n",cur_pos_rel+cue);
#endif

	switch(speed++) {
	case 10:
	  cue=cue<<1;
	  break;
	case 25:
	  cue=cue<<1;
	  break;
	case 40:
	  cue=cue<<1;
	  break;
	}

	tmppos = cur_pos_rel + cue;
	if (tmppos >= thiscd.trk[cur_track - 1].length) {
	  if (cur_track < cur_ntracks)
	    set_error(play_cd (cur_track+1, 0, cur_ntracks + 1));
	  else {
	    cue=0; speed=0;
	  }
	} else {
	  if (tmppos<0) {
	    if (cur_track > 1) {
	      int q;
	      if (cur_pos_rel<0)
		q=cur_pos_rel+tmppos;
	      else
		q=tmppos;
	      set_error(play_cd(cur_track-1,
				thiscd.trk[cur_track - 2].length+q,
				cur_ntracks + 1));
	    }
	    else {
	      cue=0; speed=0;
	    }
	  } else {
	    set_error(play_cd (cur_track, tmppos, cur_ntracks + 1));
	  }
	}
      }
    } else {

      /********    BUTTON PRESSES    ************/

      if (request<NUM_GADGETS) {
	if (request!=BTN_REWIND && request!=BTN_FF) {
	  cue=0; speed=0;
	}

	switch(request) {
	case BTN_EJECT:
	  if (checkmount()) {
	    /* nothing */
	  } else {
	    if (enable_eject) {
	      switch(set_error(eject_cd())) {
	      case 0:
		break;
	      case 2:
		/* nothing */
		break;
	      default:
		break;
	      }
	    }
	  }
	  break;
	    
	case BTN_REWIND:
	  cue=0; speed=0;

	  if (down && (cur_cdmode == CDPAUSE)) {
	    set_error(pause_cd());
	    cur_cdmode = CDPLAY;
	  }

	  if (cur_cdmode == CDPLAY) {
	    if (down) {
	      cue = -2; speed=0;
	    } else {
	      cue = 0;
	    }
	  }
	  break;

	case BTN_PLAYSTOP:
	  if (cur_cdmode == CDSTOP || cur_cdmode == CDNULL) {
	    if (cd_stopped_at_end)
	      save_track = 1;  /* If the CD has stopped at its end,
				  then start over */
	    set_error(play_cd (save_track, 0, cur_ntracks + 1));
	    break;
	  }

	  if (cur_cdmode == CDPLAY || cur_cdmode == CDPAUSE) {
	    save_track = cur_track;
	    if (save_track<1) save_track=1;
	    stop_cd ();
	    cd_stopped_at_end = 0;
	    cue=0; speed=0;
	    break;
	  }

	case BTN_PAUSE:
	  if (cur_cdmode == CDPLAY || cur_cdmode == CDPAUSE) {
	    set_error(pause_cd ());
	  }
	  break;

	case BTN_FF:
	  cue=0; speed=0;

	  if (down && (cur_cdmode == CDPAUSE)) {
	    set_error(pause_cd());
	    cur_cdmode = CDPLAY;
	  }

	  if (cur_cdmode == CDPLAY) {
	    if (down) {
	      cue = 2; speed=0;
	    } else {
	      cue = 0;
	    }
	  }
	  break;

	case BTN_SHUFFLE:
#ifdef DEBUG_CODE
	  pr_trackinfo(cur_track);
#else
	  shuffle = down;
	  UpdateNeedToPoll();
#endif
	  break;

	case BTN_REPEAT:
	  repeat = down;
	  UpdateNeedToPoll();
	  break;


	case BTN_PREV:
	  if (!down) break;
	  if (cur_cdmode == CDSTOP) {
	    save_track--;
	    if (save_track < 1)
	      save_track = 1;
	    cd_stopped_at_end = 0;
	    break;
	  }

	  if (cur_cdmode == CDPLAY || cur_cdmode == CDPAUSE) {
	    if (cur_pos_rel>2) cur_track++;
	    cur_track--;
	    if (cur_track < 1)
	      cur_track = 1;
	    else
	      set_error(play_cd (cur_track, 0, cur_ntracks + 1));
	    cd_stopped_at_end = 0;
	    break;
	  }

	  break;

	case BTN_NEXT:
	  if (!down) break;
	  if (cur_cdmode == CDSTOP) {
	    save_track++;
	    if (save_track > cur_ntracks)
	      save_track = cur_ntracks;
	    cd_stopped_at_end = 0;
	    break;
	  }
	  if (cur_cdmode == CDPLAY || cur_cdmode == CDPAUSE) {
	    cur_track++;
	    if (cur_track > cur_ntracks)
	      cur_track = cur_ntracks;
	    else
	      set_error(play_cd (cur_track, 0, cur_ntracks + 1));
	    cd_stopped_at_end = 0;
	    break;
	  }
	  break;

	case MSG_TIME:
	  clk_mode++;
	  if (clk_mode>=NUM_CLKMODES)
	    clk_mode=0;
	  break;

	}
      }
    }
  }
#ifndef EMPTY_POLL
  if (cur_cdmode==CDEJECT) enable_eject=0; else enable_eject=1;
#endif
}

/********    DISPLAY UPDATE    ************/

int cd_update() {
  int cds, min, sec;
  int flush = 0;
  char buf[256];
  int index,y;

  /* If we are not visible, no need to process further */
  if (!WindowVisible) return;

  cds=cd_status();
#ifdef DBG_STATUS
  fprintf(stderr, "Status: %d\n",cds);
#endif

  /* Statusline */
  if (error)
    switch(error) {
    case -3:
    case -2:
      strcpy(buf, "Data track");
      break;
    default:
      strcpy(buf, "Busy");
    }
  else 
    switch (cur_cdmode) {
    case CDNULL:
      strcpy(buf, "Stopped");
      break;
    case CDPLAY:
      strcpy(buf,"Playing");
      break;
    case CDPAUSE:
      strcpy(buf,"Pause");
      break;
    case CDSTOP:
      strcpy(buf,"Stopped");
      break;
    case CDEJECT:
      strcpy(buf,"No disk");
      break;
    case CDMOUNT:
      strcpy(buf,"Mounted");
      break;
    default:
      sprintf(buf,"Mode %d", cur_cdmode);
    }
  if (strcmp(Gad[MSG_STATUS]->normalLabel->text,buf)) {
    McChangeText(app, Gad[MSG_STATUS]->normalLabel, buf, 0);
    Gad[MSG_STATUS]->normalLabel->x =
      ((Gad[MSG_STATUS]->width - Gad[MSG_STATUS]->normalLabel->width)>>1) - 1;
    McGadgetRedraw(app, Gad[MSG_STATUS]);
    flush = 1;
  }

  /* update clocks */
  switch (clk_mode) {
    int pos;
  case CLK_TRACK:
    pos = (cur_pos_rel>0)?cur_pos_rel:-cur_pos_rel;
    sec = pos % 60;
    min = pos / 60;
    sprintf(buf, "%s%02d:%02d%s",
	    (cur_pos_rel<0)?"-":" ", min, sec, (cur_pos_rel<0)?"":" ");
    break;
  case CLK_ELAPSED:
    sec = cur_pos_abs % 60;
    min = cur_pos_abs / 60;
    sprintf(buf, "%02d:%02d", min, sec);
    break;
  default:
    sec = thiscd.length - cur_pos_abs;
    min = sec / 60;
    sec = sec % 60;
    sprintf(buf, "%02d:%02d", min, sec);
  }
  switch (clk_mode) {
  case CLK_TRACK:
    index = PIX_T;
    break;
  case CLK_ELAPSED:
    index = PIX_E;
    break;
  default: /* remaining */
    index = PIX_R;
    break;
  }

  if (Gad[MSG_TIME]->normalBitmap != &pmap[index]) {
    Gad[MSG_TIME]->normalBitmap = &pmap[index];
    flush = 3;
  }

  if (strcmp(Gad[MSG_TIME]->normalLabel->text,buf)) {
    McChangeText(app, Gad[MSG_TIME]->normalLabel, buf, 0);
    Gad[MSG_TIME]->normalLabel->x =
      ((Gad[MSG_TIME]->width - Gad[MSG_TIME]->normalLabel->width)>>1) + 2;
    flush = 3;
  }

  if (flush==3)
    McGadgetUpdate(app, Gad[MSG_TIME]);


  switch (cur_cdmode) {
  case CDEJECT:
    strcpy(buf, "--");
    break;
  case CDSTOP:
    sprintf(buf, "%02d", save_track);
    break;
  default:
    sprintf(buf, "%02d", cur_track);
    break;
  }
  if (strcmp(Gad[MSG_TRACK]->normalLabel->text,buf)) {
    McChangeText(app, Gad[MSG_TRACK]->normalLabel, buf, 0);
    Gad[MSG_TRACK]->normalLabel->x =
      ((Gad[MSG_TRACK]->width - Gad[MSG_TRACK]->normalLabel->width)>>1);
    McGadgetUpdate(app, Gad[MSG_TRACK]);
    flush = 1;
  }

  if (cur_cdmode == CDPLAY || cur_cdmode == CDPAUSE)
    index = PIX_STOP;
  else
    index = PIX_PLAY;
  if (Gad[BTN_PLAYSTOP]->normalBitmap != &pmap[index]) {
    Gad[BTN_PLAYSTOP]->normalBitmap = &pmap[index];
    McGadgetUpdate(app, Gad[BTN_PLAYSTOP]);
    flush = 1;
  }
    

  return flush;
}
