/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "config.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#ifdef USE_MIXER
#include <sys/soundcard.h>
#include "volume.h"
#endif

#include "McApp.h"
#include "McGadget.h"
#include "McSlider.h"
#include "McText.h"
#include "McBitmap.h"
#include "icons.h"
#include "cd.h"
#include "struct.h"

/* Private routines
 * ================
 */
static void button_proc(McGadget *gadget);
static void slider_proc(McGadget *gadget);
static void main_loop();
static void AddButton(int index);
static void AddMessage(int x, int y, int width, int txt_x, char *title);
static void AddBitmap(int index, int x, int y);
static int ScanArg(int argc, char *argv[], int i);

/* Private variables
 * =================
 */

#ifdef USE_MIXER
int use_mixer = -1; /* probe for mixer first */
#endif

/*  Public routines
 * ===============
 */

static int update_volume(int left, int right);

/* Public variables
 * ================
 */

int		screen_num;
Screen		*screen_ptr;
char		*icon_name = "XPlaycd";
McGadget	*Gad[NUM_GADGETS];
McApp		*app = NULL;
char		window_title[128]="XPlaycd", *window_name = window_title;
char		*xplaycd_version = "XPlaycd 1.1";

int             WindowHeight;
int             WindowVisible = 0;
int             NeedToPoll = 0;
int             GadCnt;
char		*myname;

/**************************************************************************
 *  main(int ac, char **argv)
 *
 */
 main(int argc, char **argv) {
  int	i,j, left, right;

  myname = argv[0];

  /*
   * Connect to the cdrom
   */
  cd_init();

  /*
   * Build connection with X server
   */
  app = McCreateApp(McConnectToServer(&argc, &argv, ScanArg));

  if (argc>1) {
    fprintf(stderr,"%s: Warning: Garbage on command line.\n",myname);
  }

  /*
   * Create the window
   */
  WindowHeight=106;
  McCreateAppWindow(app, 360, WindowHeight);

  /*
   * Create all bitmaps
   */

  for (i=NUM_BITMAPS; --i>=0;) {
    pmap[i].pixmap = XCreateBitmapFromData(app->display, app->window,
					   pmap[i].bits, pmap[i].width,
					   pmap[i].height);
  }

  /*
   * Set WN Hints
   */
 {
   XWMHints	 wm_hints;
   XClassHint	 class_hints;
   XTextProperty windowName;
   XTextProperty iconName;
   XSizeHints	 sizehints;

   if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
     fprintf(stderr, "%s: structure allocation for windowName failed.\n",
	     myname);
     exit(1);
   }
   if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
     fprintf(stderr, "%s: structure allocation for iconName failed.\n",
	     myname);
     exit(1);
   }

   wm_hints.initial_state = NormalState;
   wm_hints.input = True;
   wm_hints.icon_pixmap = pmap[PIX_ICON].pixmap;
   wm_hints.icon_mask = pmap[PIX_ICONMASK].pixmap;
   wm_hints.flags = StateHint | IconPixmapHint | IconMaskHint | InputHint;
   class_hints.res_name = myname;
   class_hints.res_class = "Multimedia";
   sizehints.flags = PSize | PMinSize | PMaxSize | PResizeInc;
   sizehints.width = sizehints.min_width = sizehints.max_width = 360;
   sizehints.height= sizehints.min_height= WindowHeight;
   sizehints.max_height = 400;
   sizehints.width_inc = sizehints.height_inc = 1;
   
   XSetWMProperties(app->display, app->window, &windowName, &iconName,
		    argv, argc, &sizehints, &wm_hints, &class_hints);
 }

  /*
   * Select the input events we will be wanting to deal with
   */
  XSelectInput(app->display, app->window, 
	       ExposureMask |
	       KeyPressMask |
	       VisibilityChangeMask |
	       ButtonPressMask |
	       ButtonReleaseMask |
	       ButtonMotionMask |
	       StructureNotifyMask);

  /*
   * Display the window
   */
  XMapWindow(app->display, app->window);

  GadCnt=1;
  Gad[0] = McCreateGadget(app, GAD_H3D | GAD_3D | GAD_ACTIVE, BOOLGADGET,
			  10, 11, 39, 20);
  Gad[0]->normalLabel = McCreateText(app, "Exit", app->gc[GC_NORMAL],
				     app->normalFont, 8, 4);
  Gad[0]->callbackUp = button_proc;
  Gad[0]->id = 0;
  AddButton(PIX_EJECT);
  AddButton(PIX_REWIND);
  Gad[GadCnt-1]->callbackUp = Gad[GadCnt-1]->callbackDown;
  AddButton(PIX_PLAY);
  AddButton(PIX_FF);
  Gad[GadCnt-1]->callbackUp = Gad[GadCnt-1]->callbackDown;
  AddButton(PIX_PAUSE);
  AddButton(PIX_SHUFFLE);
#ifndef DEBUG_CODE
  Gad[GadCnt-1]->flags |= GAD_TOGGLE | GAD_DISABLE;
  Gad[GadCnt-1]->flags &= ~GAD_ACTIVE; /* Not implemented yet */
#endif
  AddButton(PIX_REPEAT);
  Gad[GadCnt-1]->flags |= GAD_TOGGLE;
  AddButton(PIX_PREV);
  Gad[GadCnt-1]->x = Gad[BTN_REWIND]->x;
  Gad[GadCnt-1]->y = 40;
  AddButton(PIX_NEXT);
  Gad[GadCnt-1]->x = Gad[BTN_FF]->x;
  Gad[GadCnt-1]->y = 40;

  AddMessage( 34, 40, 57, 12, "00:00");
  Gad[GadCnt-1]->flags |= GAD_ACTIVE | GAD_HNONE | GAD_3D | GAD_H3D;
  Gad[GadCnt-1]->callbackDown = button_proc;
  Gad[GadCnt-1]->normalBitmap = &pmap[PIX_T];

  AddMessage(143, 40, 32,  8,    "00");
  AddMessage(227, 40, 67,  8, "ready");

  AddBitmap(PIX_CDLOGO, 299, 60);
  AddBitmap(PIX_VOLUME,   9, 73);
  AddBitmap(PIX_CLOCK,    9, 43);

  Gad[GadCnt] = McCreateGadget(app,
			       GAD_3D | GAD_H3D | GAD_ACTIVE | GAD_SELECTED,
			       SLIDERGADGET, 34, 69, 260, 25);
  Gad[GadCnt]->specialInfo = McCreateSlider(SLIDER_STEREO, 0, 100);
  ((McSlider *)(Gad[GadCnt]->specialInfo))->step = 3;
  Gad[GadCnt]->callbackDown = slider_proc;
  Gad[GadCnt]->id = GadCnt++;

  cd_update();

#ifdef USE_MIXER
  if (use_mixer>0)
    if (mix_open()<0) {
      fprintf(stderr,"%s: Can't open " MIXERDEVICE ": ",myname);
      perror("");
      cleanup(1);
    }

  if (use_mixer<0)
    if (mix_open()>=0) use_mixer=1; else use_mixer=0;

  if (use_mixer) {
    left=right=-1;
    mix_set_volume(SOUND_MIXER_CD, &left, &right);
  } else {
    cd_status();
    left=right=50;
    cd_volume(left, right);
  }
#else
    cd_status();
    left=right=50;
    cd_volume(left, right);
#endif
  update_volume(left, right);
  /*
   * Now do the main loop
   */
  main_loop();
}

static int ScanArg(int argc, char *argv[], int i) {
  if ((!strcmp(argv[i], "-mixer")) || (!strcmp(argv[i], "-m"))) {
#ifdef USE_MIXER
    use_mixer=1;
    return 1;
#else
    fprintf(stderr,"%s: Compiled without mixer support.\n",myname);
    cleanup(1);
#endif
  }
  if ((!strcmp(argv[i], "-nomixer")) || (!strcmp(argv[i], "-nm"))) {
#ifdef USE_MIXER
    use_mixer=0;
    return 1;
#endif
  }
  if (!strcmp(argv[i], "-v")) {
    printf("%s\n", xplaycd_version);
    cleanup(0);
  }
  return 0;
}

/**************************************************************************
 * static void button_proc(McGadget *gadget)
 *
 */
static void button_proc(McGadget *gadget) {
  if (gadget->id == BTN_EXIT)
    cleanup(0);
  cd_action(gadget->id, gadget->flags&GAD_SELECTED);
  cd_update(); /* update messages */
}

/**************************************************************************
 * VOLUME VOLUME VOLUME VOLUME VOLUME VOLUME VOLUME VOLUME VOLUME VOLUME */

#define SLIDER(gad) ((McSlider *)((gad)->specialInfo))

static void slider_proc(McGadget *gadget) {
  int left, right;
  left = SLIDER(Gad[SLD_VOLUME])->leftValue;
  right = SLIDER(Gad[SLD_VOLUME])->rightValue;
#ifdef USE_MIXER
  if (use_mixer) {
    mix_set_volume(SOUND_MIXER_CD, &left, &right);
    update_volume(left, right);
  } else {
    cd_volume(left, right);
  }
#else
  cd_volume(left, right);
#endif
}

static int update_volume(int left, int right) {
  if ((left>=0) && (right>=0) &&
      ((left != (SLIDER(Gad[SLD_VOLUME])->leftValue)) ||
       (right != (SLIDER(Gad[SLD_VOLUME])->rightValue)))) {
    SLIDER(Gad[SLD_VOLUME])->leftValue = left;
    SLIDER(Gad[SLD_VOLUME])->rightValue = right;
    McGadgetUpdate(app, Gad[SLD_VOLUME]);
    return 1;
  }
  return 0;
}

/**************************************************************************
 * static void main_loop()
 *
 */

void myXNextEvent(Display *display, XEvent *event_return) {
  struct timeval to;
  fd_set rdfds, usefds;
  int dfd, width, flush;

  if (XPending(display)>0) {
    cd_update(); /* update messages */
    XNextEvent(display, event_return);
    return;
  }

  FD_ZERO(&rdfds);
  dfd=ConnectionNumber(display);
  width = dfd;
  FD_SET(dfd,&rdfds);

  while(1) {
    flush=0;
    if (cue)
      to.tv_usec = 100000;  /* Poll faster while searching */
    else
      if (WindowVisible)
	to.tv_usec =  333333; /* three polls per second should be enough to
			       * have a smooth display */
      else
	to.tv_usec = 1000000; /* When the window is iconified, we only have
			       * to poll to ensure repeat and shuffle
			       * functions. Therefore, once per second is
			       * enough then. */

    to.tv_sec  = 0;
    usefds=rdfds;
#ifdef EMPTY_POLL
    if (WindowVisible || NeedToPoll) {
#else
    if ((WindowVisible || NeedToPoll) && (cur_cdmode!=CDEJECT)) {
#endif
      select(width+1, &usefds, (fd_set *) NULL, (fd_set *) NULL, &to);
    } else {
      select(width+1, &usefds, (fd_set *) NULL, (fd_set *) NULL,
	     (struct timeval *) NULL);
    }    

    if (!FD_ISSET(dfd, &usefds)) {
      cd_action(REQ_POLL, 0);
      flush+=cd_update();
#ifdef USE_MIXER
      if (use_mixer) {
	int l,r;
	mix_get_volume(SOUND_MIXER_CD, &l, &r);
	flush+=update_volume(l,r);
      }
#endif
    }
    if (XPending(display)>0) {
      XNextEvent(display, event_return);
      return;
    }
    if (flush) XFlush(app->display);
  }
}

static void main_loop() {
  XEvent report;
  int i;

  while(1) {
    myXNextEvent(app->display, &report);
    switch (report.type) {
    case UnmapNotify:
      WindowVisible = 0;
      break;
    case MapNotify:
      WindowVisible = 1;
      break;
    default:
      McAppDoEvent(app, &report);
      break;
    }
  }
}

/**************************************************************************
 * GADGETS  GADGETS  GADGETS  GADGETS  GADGETS  GADGETS  GADGETS  GADGETS */

static void AddButton(int index) {
  Gad[GadCnt] = McCreateGadget(app, GAD_3D | GAD_H3D | GAD_ACTIVE, BOOLGADGET,
			       Gad[GadCnt-1]->x + Gad[GadCnt-1]->width + 10,
			       Gad[GadCnt-1]->y, 32, 20);
  Gad[GadCnt]->callbackDown = button_proc;
  Gad[GadCnt]->normalBitmap = &pmap[index];
  Gad[GadCnt]->id = GadCnt++;
}


static void AddMessage(int x, int y, int width, int txt_x, char *title) {
  Gad[GadCnt] = McCreateGadget(app, GAD_3D | GAD_H3D | GAD_SELECTED,
			       BOOLGADGET, x, y, width, 20);
  Gad[GadCnt]->normalLabel = McCreateText(app, title, app->gc[GC_NORMAL],
					  app->normalFont, txt_x, 3);
  Gad[GadCnt]->id = GadCnt++;
}


static void AddBitmap(int index, int x, int y) {
  Gad[GadCnt] = McCreateGadget(app, GAD_HNONE, BOOLGADGET, x, y,
			       pmap[index].width, pmap[index].height);
  Gad[GadCnt]->normalBitmap = &pmap[index];
  Gad[GadCnt]->id = GadCnt++;
}

/**************************************************************************
 * BE CLEAN  BE CLEAN  BE CLEAN  BE CLEAN  BE CLEAN  BE CLEAN  BE CLEAN   */

cleanup(r) {
  if (app) {
    if (app->display)
      XCloseDisplay(app->display);
    free(app);
  }
#ifdef USE_MIXER
  mix_close();
#endif
  exit(r);
}

