/**********************************************************************/
/*                                                                    */
/*  File:          xvumeter.c                                         */
/*                                                                    */
/**********************************************************************/

#include "config.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <linux/soundcard.h>

#include "McApp.h"
#include "McGadget.h"
#include "McSlider.h"
#include "McText.h"
#include "McBitmap.h"
#include "xvumeter.h"


char *xvumeter_version;
char *myname;
char *window_name, *icon_name;
McApp *app;
int WindowWidth, WindowHeight;
int WindowVisible;

int GadCnt;
McGadget *Gad[NUM_GADGETS];

static int ScanArg(int argc, char *argv[], int i);
static void exit_proc(McGadget *gadget);
static void main_loop();




main(int argc, char **argv) {

  myname = argv[0];
  window_name = icon_name = "XVumeter";
  xvumeter_version = "XVumeter V1.0";

  /*
   * Build connection with X server
   */
  app = McCreateApp(McConnectToServer(&argc, &argv, ScanArg));
  if (argc>1) {
    fprintf(stderr,"%s: Warning: Garbage on command line.\n",myname);
  }

  WindowWidth  = 300;
  WindowHeight = 90;

  /*
   * Create the window
   */
  McCreateAppWindow(app, WindowWidth, WindowHeight);

  /*
   * Set WM Hints
   */
 {
   XWMHints	 wm_hints;
   XClassHint	 class_hints;
   XTextProperty windowName;
   XTextProperty iconName;
   XSizeHints	 sizehints;

   if (XStringListToTextProperty(&window_name, 1, &windowName) == 0) {
     fprintf(stderr, "%s: structure allocation for windowName failed.\n",
	     argv[0]);
     cleanup(1);
   }
   if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
     fprintf(stderr, "%s: structure allocation for iconName failed.\n",
	     argv[0]);
     cleanup(1);
   }

   wm_hints.initial_state = NormalState;
   wm_hints.input = True;
   wm_hints.flags = StateHint | InputHint;
   class_hints.res_name = argv[0];
   class_hints.res_class = "XVumeter";
   sizehints.flags = PSize | PMinSize | PMaxSize | PResizeInc;
   sizehints.min_width = 0;
   sizehints.width = sizehints.max_width = WindowWidth;
   sizehints.height= sizehints.max_height= WindowHeight;
   sizehints.min_height= 0;
   sizehints.width_inc = sizehints.height_inc = 1;
   
   XSetWMProperties(app->display, app->window, &windowName, &iconName,
		    argv, argc, &sizehints, &wm_hints, &class_hints);
 }

  /*
   * Select the input events we will be wanting to deal with
   */
  XSelectInput(app->display, app->window, 
	       ExposureMask |
	       KeyPressMask |
	       ButtonPressMask |
	       ButtonReleaseMask |
	       ButtonMotionMask |
	       StructureNotifyMask);

  /*
   * Display the window
   */
  XMapWindow(app->display, app->window);
  XFlush(app->display);


  GadCnt=1;
  Gad[0] = McCreateGadget(app, GAD_H3D | GAD_3D | GAD_ACTIVE, BOOLGADGET,
			  10, 11, 39, 20);
  Gad[0]->normalLabel = McCreateText(app, "Exit", app->gc[GC_NORMAL],
				     app->normalFont, 8, 4);
  Gad[0]->callbackUp = exit_proc;

  Gad[1] = McCreateGadget(app,
			  GAD_3D | GAD_H3D | GAD_SELECTED,
			  SLIDERGADGET, 10,40,255,20);
  Gad[1]->specialInfo = McCreateSlider(SLIDER_MONO, 0, 127);
  GadCnt=2;

  main_loop();
}

static int ScanArg(int argc, char *argv[], int i) {

  if (!strcmp(argv[i], "-v")) {
    printf("%s\n", xvumeter_version);
    cleanup(0);
  }
  return 0;
}

/**************************************************************************
 * static void exit_proc(McGadget *gadget)
 *
 */
static void exit_proc(McGadget *gadget) {
  cleanup(0);
}

cleanup(int r) {
  if (app) {
    if (app->display)
      XCloseDisplay(app->display);
    free(app);
  }
  exit(r);
}

/**************************************************************************
 * static void main_loop()
 *
 */

#define BUFSIZE 512

void myXNextEvent(Display *display, XEvent *event_return) {
  struct timeval to;
  fd_set rdfds, usefds;
  int dfd;
  int num, upd, val, max, med, cnt, dec;
  unsigned char buf[BUFSIZE], *pt;

  if (XPending(display)>0) {
    XNextEvent(display, event_return);
    return;
  }

  FD_ZERO(&rdfds);
  dfd=ConnectionNumber(display);
  FD_SET(dfd,&rdfds);
  FD_SET(fileno(stdin),&rdfds);
  to.tv_usec = 0; to.tv_sec=0;
  cnt = 0; max=0; dec=2;

  while(1) {
    usefds=rdfds;
    if ((to.tv_sec==0) && (to.tv_usec==0)) {
      to.tv_usec = 40000; to.tv_sec=0;
      if (max) {
	if (cnt>10) {
	  max-=dec;
	  upd=1;
	  if (max<0) max=0;
	}
	if (++cnt>25) { dec=dec+dec; cnt=5; }
      }
      if (upd) {
	((McSlider *)(Gad[SLI_VU]->specialInfo))->leftValue = max;
	McGadgetUpdate(app, Gad[SLI_VU]);
	upd=0;
      }
    }
    select(dfd+1, &usefds, (fd_set *) NULL, (fd_set *) NULL, &to);

    if (FD_ISSET(fileno(stdin),&usefds)) {
      if ((num=read(fileno(stdin),buf,BUFSIZE))>0) {
	for(pt=buf;--num>=0;pt++) {
	  if ((val=*pt)>=128) val-=128; else val=127-val; /* AC --> DC */
	  if (val>max) { max=val; upd=1; dec=2; cnt=0; }
	}
      }
    }

    if (XPending(display)>0) {
      XNextEvent(display, event_return);
      return;
    }
  }
}

static void main_loop() {
  XEvent report;
  int i;

  while(1) {
    myXNextEvent(app->display, &report);
    switch (report.type) {
    case Expose:
      McAppDoEvent(app, &report);
      break;
    case UnmapNotify:
      WindowVisible = 0;
      break;
    case MapNotify:
      WindowVisible = 1;
      break;
    default:
      McAppDoEvent(app, &report);
      break;
    }
  }
}
