/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/bitmaps/gray1>
#include <X11/bitmaps/gray3>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <varargs.h>

#include "McApp.h"
#include "McGadget.h"

static char *colors[] = COLORS;
char *display_name;
char *normal_font_name;
char *bold_font_name;
char *myname;

/* STARTUP STARTUP STARTUP STARTUP STARTUP STARTUP STARTUP STARTUP STARTUP */

Display *McConnectToServer(int *ac, char **av[],
			   int (*parg)(int argc, char *argv[], int i)) {
  int i, j, skip, argc;
  char **argv;
  Display *display;

  argc = *ac;
  argv = *av;
  display_name = NULL;
  normal_font_name = DEFAULT_FONT;
  bold_font_name = BOLD_FONT;
  myname = argv[0];

  skip = 0;
  i=1;
  while (i<argc) {

    if (skip) {
      for (j=i+skip;j<argc;j++) argv[j-skip]=argv[j];
      argc-=skip;
      skip = 0;
    }
    if (i>=argc) break;

    if (strcmp(argv[i], "-display") == 0) {
      if ((i+1) < argc) {
	display_name=argv[i+1];
	skip = 2; continue;
      } else {
	fprintf(stderr,"%s: Missing argument\n",myname);
	exit(1);
      }
    }
    if (strcmp(argv[i], "-fn") == 0) {
      if ((i+1) < argc) {
	normal_font_name=argv[i+1];
	skip = 2; continue;
      } else {
	printf("%d %d\n", i, argc);
	fprintf(stderr,"%s: Missing argument\n",myname);
	exit(1);
      }
    }
    if (strcmp(argv[i], "-bfn") == 0) {
      if ((i+1) < argc) {
	bold_font_name=argv[i+1];
	skip = 2; continue;
      } else {
	fprintf(stderr,"%s: Missing argument\n",myname);
	exit(1);
      }
    }

    if (!(parg && (skip = (parg)(argc,argv,i))))
      i++;
  }

  *ac = argc;
  *av = argv;

  /*
   * Build connection with X server
   */
  if ((display = XOpenDisplay(display_name)) == NULL) {
    fprintf(stderr, "%s: Cannot connect to X server %s\n", 
	    myname, XDisplayName(display_name));
    exit(1);
  }
  return display;
}

McApp *McCreateApp(Display *display) {
  XGCValues values;
  XColor color, rcolor;
  int i;
  McApp *app;

  if (!(app = (McApp *)malloc(sizeof(McApp))))
    return NULL;

  app->display = display;

  app->screen = DefaultScreen(display);

  if (!(app->normalFont = XLoadQueryFont(display, normal_font_name))) {
    fprintf(stderr,"%s: Can't open font '%s', using 'fixed'.\n",
	    myname, normal_font_name);
    if (!(app->normalFont = XLoadQueryFont(display, "fixed"))) {
      fprintf(stderr,"%s: Can't open font 'fixed' either!\n", myname);
      exit(1);
    }
  }

  if (!(app->boldFont = XLoadQueryFont(display, bold_font_name))) {
    fprintf(stderr,"%s: Can't open font '%s', using 'fixed'.\n",
	    myname, bold_font_name);
    if (!(app->boldFont = XLoadQueryFont(display, "fixed"))) {
      fprintf(stderr,"%s: Can't open font 'fixed' either!\n", myname);
      exit(1);
    }
  }

  app->firstGadget = app->lastGadget = NULL;

  app->stipple1 = XCreateBitmapFromData(display,
					RootWindow(display, app->screen),
					gray1_bits, gray1_width, gray1_height);

  app->stipple3 = XCreateBitmapFromData(display,
					RootWindow(display, app->screen),
					gray3_bits, gray3_width, gray3_height);


  if (DefaultDepth(display, app->screen) > 1) {

    app->hasColor = True;

    app->colormap = DefaultColormap(app->display, app->screen);

    for (i = 0; i < MAX_COLORS; i++) {

      XAllocNamedColor(app->display, app->colormap, colors[i],
		       &color, &rcolor);

      app->colors[i] = rcolor.pixel;
    }

    values.foreground = app->colors[COL_FOREGROUND];

    values.background = app->colors[COL_BACKGROUND];

    values.font       = app->normalFont->fid;

    values.fill_style = FillStippled;

    values.stipple    = app->stipple3;


  for (i=GC_NORMAL;i<=GC_REVERSE;i++) {
    app->gc[i] = XCreateGC(app->display,
			  RootWindow(app->display, app->screen),
			  GCForeground | GCBackground | GCFont |
			  GCStipple, &values);
  }
    app->gc[GC_BUSY] = XCreateGC(app->display,
				RootWindow(app->display, app->screen),
				GCForeground | GCBackground | GCFont |
				GCFillStyle | GCStipple, &values);
    
    for (i=GC_SELECTED;i<=GC_SELECTED_BITMAP;i++) {
      app->gc[i] = XCreateGC(app->display,
			    RootWindow(app->display, app->screen),
			    GCForeground | GCBackground | GCFont, &values);
    }

    XSetForeground(app->display, app->gc[GC_REVERSE],
		   app->colors[COL_BACKGROUND]);
    XSetBackground(app->display, app->gc[GC_REVERSE],
		   app->colors[COL_FOREGROUND]);
    XSetForeground(app->display, app->gc[GC_BRIGHT],
		   app->colors[COL_BRIGHT]);
    XSetForeground(app->display, app->gc[GC_DARK],
		   app->colors[COL_DARK]);
    XSetForeground(app->display, app->gc[GC_SELECTED],
		   app->colors[COL_SELECTED]);
    XSetForeground(app->display, app->gc[GC_SELECTED_BITMAP],
		   app->colors[COL_FOREGROUND]);
    XSetBackground(app->display, app->gc[GC_SELECTED_BITMAP],
		   app->colors[COL_SELECTED]);
  } else { /* mono */
    app->hasColor = False;
    values.foreground = BlackPixel(app->display, app->screen);
    values.background = WhitePixel(app->display, app->screen);
    values.font = app->normalFont->fid;
    values.stipple = app->stipple3;

    app->gc[GC_NORMAL] = XCreateGC(app->display,
				  RootWindow(app->display, app->screen),
				  GCForeground | GCBackground | GCFont |
				  GCStipple, &values);
    values.fill_style = FillStippled;
    app->gc[GC_BUSY] = XCreateGC(app->display,
				RootWindow(app->display, app->screen),
				GCForeground | GCBackground | GCFont |
				GCFillStyle | GCStipple, &values);
    values.stipple = app->stipple1;
    values.fill_style = FillOpaqueStippled;
    app->gc[GC_BRIGHT] = XCreateGC(app->display,
				  RootWindow(app->display, app->screen),
				  GCForeground | GCBackground | GCFont |
				  GCFillStyle | GCStipple, &values);
    app->gc[GC_SELECTED_BITMAP]=app->gc[GC_DARK] = XCreateGC(app->display,
				RootWindow(app->display, app->screen),
				GCForeground | GCBackground | GCFont,
				&values);
    app->gc[GC_SELECTED]=XCreateGC(app->display,
				    RootWindow(app->display, app->screen),
				    GCForeground | GCBackground | GCFont,
				    &values);
    XSetForeground(app->display, app->gc[GC_SELECTED], values.background);
    XSetBackground(app->display, app->gc[GC_SELECTED], values.foreground);
    app->gc[GC_REVERSE] = XCreateGC(app->display,
				   RootWindow(app->display, app->screen),
				   GCForeground | GCBackground | GCFont |
				   GCStipple, &values);
    XSetForeground(app->display, app->gc[GC_REVERSE], values.background);
    XSetBackground(app->display, app->gc[GC_REVERSE], values.foreground);
  }

  /* both, color & mono */
  app->borderWidth = DEFAULT_BORDER_WIDTH;
  XSetFunction(app->display, app->gc[GC_BUSY], GXcopy);

  return app;
}


Window McCreateAppWindow(McApp *app, int width, int height) {
  Window win;
  if (app->hasColor) {
    win = XCreateSimpleWindow(app->display,
			      RootWindow(app->display, app->screen),
			      0, 0, width, height, 0,
			      app->colors[COL_FOREGROUND],
			      app->colors[COL_BACKGROUND]);
  } else {
    win = XCreateSimpleWindow(app->display,
			      RootWindow(app->display, app->screen),
			      0, 0, width, height, 0,
			      BlackPixel(app->display, app->screen),
			      WhitePixel(app->display, app->screen));
  }
  return app->window = win;
}

/* 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D 3D */

void McAppDrawbox(McApp *app, int x, int y, int width, int height, int mode) {
  int i;
  int thick = app->borderWidth;
  XPoint points[6];

  switch (mode) {
  case _3D_OUT:
    points[0].x = x - thick;		points[0].y = y - thick;
    points[1].x = width + 2 * thick;	points[1].y = 0;
    points[2].x = -thick;		points[2].y = thick;
    points[3].x = -width;		points[3].y = 0;
    points[4].x = 0;			points[4].y = height;
    points[5].x = -thick;		points[5].y = thick;
    XFillPolygon(app->display, app->window, app->gc[GC_BRIGHT],
		 points, 6, Nonconvex, CoordModePrevious);
    points[0].y = y + height + thick;
    points[1].x = thick;		points[1].y = -thick;
    points[2].x = width;		points[2].y = 0;
    points[3].x = 0;			points[3].y = -height;
    points[4].x = thick;		points[4].y = -thick;
    points[5].x = 0;			points[5].y = height + 2 * thick;
    XFillPolygon(app->display, app->window, app->gc[GC_DARK],
		 points, 6, Nonconvex, CoordModePrevious);
    break;

  case _3D_IN:
    points[0].x = x - thick;		points[0].y = y - thick;
    points[1].x = width + 2 * thick;	points[1].y = 0;
    points[2].x = -thick;		points[2].y = thick;
    points[3].x = -width;		points[3].y = 0;
    points[4].x = 0;			points[4].y = height;
    points[5].x = -thick;		points[5].y = thick;
    XFillPolygon(app->display, app->window, app->gc[GC_DARK],
		 points, 6, Nonconvex, CoordModePrevious);
    points[0].y = y + height + thick;
    points[1].x = thick;		points[1].y = -thick;
    points[2].x = width;		points[2].y = 0;
    points[3].x = 0;			points[3].y = -height;
    points[4].x = thick;		points[4].y = -thick;
    points[5].x = 0;			points[5].y = height + 2 * thick;
    XFillPolygon(app->display, app->window, app->gc[GC_BRIGHT],
		 points, 6, Nonconvex, CoordModePrevious);
    break;
    
  case _3D_FLAT:
    points[0].x = x - thick;		points[0].y = y - thick;
    points[1].x = width + 2 * thick;	points[1].y = 0;
    points[2].x = -thick;		points[2].y = thick;
    points[3].x = -width;		points[3].y = 0;
    points[4].x = 0;			points[4].y = height;
    points[5].x = -thick;		points[5].y = thick;
    XFillPolygon(app->display, app->window, app->gc[GC_REVERSE],
		 points, 6, Nonconvex, CoordModePrevious);
    points[0].y = y + height + thick;
    points[1].x = thick;		points[1].y = -thick;
    points[2].x = width;		points[2].y = 0;
    points[3].x = 0;			points[3].y = -height;
    points[4].x = thick;		points[4].y = -thick;
    points[5].x = 0;			points[5].y = height + 2 * thick;
    XFillPolygon(app->display, app->window, app->gc[GC_REVERSE],
		 points, 6, Nonconvex, CoordModePrevious);
    break;
  }
}

int McInRectangle(int ex, int ey, int x, int y, int width, int height) {
  if (ex >= x && ex < x + width &&
      ey >= y && ey < y + height)
    return 1;
  else
    return 0;
}

/* LISTS LISTS LISTS LISTS LISTS LISTS LISTS LISTS LISTS LISTS LISTS LISTS */

void McAddGadgetToList(McApp *app, McGadget *gadget) {
  if (!(app->firstGadget))
    app->firstGadget = gadget;

  if (!(app->lastGadget)) {
    app->lastGadget = gadget;
    gadget->prev = gadget->next = NULL;
  } else {
    gadget->prev = app->lastGadget;
    gadget->next = NULL;
    app->lastGadget->next = gadget;
    app->lastGadget = gadget;
  }
}

void McRemoveGadgetFromList(McApp *app, McGadget *gadget) {
  if (gadget->prev)
    gadget->prev->next=gadget->next;
  if (gadget->next)
    gadget->next->prev=gadget->prev;

  if (app->firstGadget == gadget)
    app->firstGadget = gadget->next;
  if (app->lastGadget == gadget)
    app->lastGadget = gadget->prev;

  gadget->prev = gadget->next = NULL;

}    

void McMoveGadgetToStart(McApp *app, McGadget *gadget) {

  McRemoveGadgetFromList(app, gadget);
  if (!(app->lastGadget))
    app->lastGadget = gadget;

  if (!(app->firstGadget)) {
    app->firstGadget = gadget;
    gadget->prev = gadget->next = NULL;
  } else {
    gadget->next = app->firstGadget;
    gadget->prev = NULL;
    app->firstGadget->prev = gadget;
    app->firstGadget = gadget;
  }
}

/* EVENTS EVENTS EVENTS EVENTS EVENTS EVENTS EVENTS EVENTS EVENTS EVENTS */

/* returns 1 if event was eaten */
int McAppDoEvent(McApp *app, XEvent *event) {
  McGadget *next;
  next = app->firstGadget;
  while(next) {
    if (McGadgetEvent(app, next,event)) return 1;
    next = next->next;
  }
  return 0;
}

void McFreeApp(McApp *app) {
  if (app) {
    if (app->normalFont) XFree(app->normalFont);
    if (app->boldFont) XFree(app->boldFont);
    XCloseDisplay(app->display);
    free(app);
  }
}
