/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _McApp_h_
#define _McApp_h_
#include "../config.h"

/* The various drawing modes for the 3d effect */

enum {
  _3D_OUT,
  _3D_IN,
  _3D_FLAT,
  _3D_NONE,
};

/* The GC's are stored in an array inside the McApp structure */
enum {
  GC_NORMAL,
  GC_REVERSE,
  GC_BUSY,
  GC_SELECTED,
  GC_BRIGHT,
  GC_DARK,
  GC_SELECTED_BITMAP,

  NUM_GC,
};

/* Theese are our colors, they are fixed for now */

#define COLORS { "gray80", "black", "gray94", "gray40", "gray72", "white", "black" };

enum {
  COL_BACKGROUND,
  COL_FOREGROUND,
  COL_BRIGHT,
  COL_DARK,
  COL_SELECTED,
  COL_WHITE,
  COL_BLACK,

  MAX_COLORS,
};

struct McApp;
struct McBitmap;
struct McText;
struct McGadget;

typedef struct McApp {
  Display *display;
  int screen;
  int borderWidth;
  Window window;
  GC gc[NUM_GC];
  XFontStruct *normalFont;
  XFontStruct *boldFont;
  Pixmap stipple1;
  Pixmap stipple3;
  int hasColor;
  Colormap colormap;
  int colors[MAX_COLORS];
  struct McGadget *firstGadget;
  struct McGadget *lastGadget;
} McApp;

#define	DEFAULT_BORDER_WIDTH	2

extern Display *McConnectToServer(int *ac, char **av[],
				  int (*parg)(int argc, char *argv[], int i));
extern McApp *McCreateApp(Display *display);
extern void McAppDrawbox(McApp *app,int x,int y,int width,int height,int mode);
extern int McInRectangle(int ex, int ey, int x, int y, int width, int height);
Window McCreateAppWindow(McApp *app, int width, int height);

extern void McAddGadgetToList(McApp *app, struct McGadget *gadget);
extern void McRemoveGadgetFromList(McApp *app, struct McGadget *gadget);
extern void McMoveGadgetToStart(McApp *app, struct McGadget *gadget);

extern int McAppDoEvent(McApp *app, XEvent *event);

extern void McFreeApp(McApp *app);

extern char *display_name;
extern char *normal_font_name;
extern char *bold_font_name;
extern char *myname;

#endif /* _McApp_h_ */
