/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <X11/Xlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "McApp.h"
#include "McGadget.h"
#include "McSlider.h"

static int BoolGadgetEvent(McApp *app, McGadget *gadget, XEvent *event);
static void GadgetUpdate(McApp *app, McGadget *gadget, int busy);

McGadget *McCreateGadget(McApp *app, unsigned int flags, unsigned int type,
			 int x, int y, int w, int h) {
  McGadget *gadget = (McGadget *) malloc(sizeof(McGadget));
  gadget->type = type;
  gadget->flags = flags;
  gadget->normalBitmap = gadget->selectBitmap = NULL;
  gadget->normalLabel = gadget->selectLabel = NULL;
  gadget->callbackDown = gadget->callbackUp = NULL;
  gadget->x = x;
  gadget->y = y;
  gadget->width = w;
  gadget->height = h;
  gadget->id = 0;
  gadget->customData = gadget->specialInfo = NULL;
  McAddGadgetToList(app, gadget);
  return gadget;
}


static void GadgetRedraw(McApp *app, McGadget *gadget, int busy) {
  int i;
  McText *txt;
  McBitmap *bmp;

  if (gadget->flags & GAD_INVIS) return;
  if (gadget->flags & GAD_STATE) {  /* Highlighted */
    if (gadget->flags & GAD_3D) {
      i = _3D_IN;
    } else {
      switch (gadget->flags & GAD_HMASK) {
      case GAD_H3D:
	i = _3D_IN;
	break;
      default:
	i = _3D_NONE;
	break;
      }
    }

    if (!(gadget->flags & GAD_NOFILL))
      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x, gadget->y, gadget->width, gadget->height);
    McAppDrawbox(app, gadget->x, gadget->y, gadget->width, gadget->height, i);

  } else { /* Normal */
    if ((((gadget->flags & GAD_HMASK) != GAD_HNONE) ||
	(gadget->normalLabel)) && (!(gadget->flags & GAD_NOFILL)))
      XClearArea(app->display, app->window,
		 gadget->x, gadget->y, gadget->width, gadget->height, False);
    if (gadget->flags & GAD_3D)
      McAppDrawbox(app, gadget->x, gadget->y, gadget->width, gadget->height,
		   _3D_OUT);
  }

  GadgetUpdate(app, gadget, busy);

  if (busy || (gadget->flags & GAD_DISABLE)) {
    XFillRectangle(app->display, app->window, app->gc[GC_BUSY],
		   gadget->x, gadget->y, gadget->width, gadget->height);
  }
}

static void GadgetUpdate(McApp *app, McGadget *gadget, int busy) {
  McText *txt;
  McBitmap *bmp;

  if (gadget->flags & GAD_INVIS) return;
  if (gadget->flags & GAD_STATE) {  /* Highlighted */
    if (gadget->specialInfo)
      (*(gadget->specialInfo->updateProc))(app, gadget, busy);
    if ((bmp=gadget->selectBitmap) || (bmp=gadget->normalBitmap))
      McPutBitmap(app, bmp, app->gc[GC_SELECTED_BITMAP], gadget->x, gadget->y);
    if ((txt=gadget->selectLabel) || (txt=gadget->normalLabel))
      McWriteText(app, txt, app->gc[GC_SELECTED_BITMAP], gadget->x, gadget->y);
  } else {
    if (gadget->specialInfo)
      (*(gadget->specialInfo->updateProc))(app, gadget, busy);
    if (gadget->normalBitmap)
      McPutBitmap(app, gadget->normalBitmap, app->gc[GC_NORMAL],
		  gadget->x, gadget->y);
    if (gadget->normalLabel)
      McWriteText(app, gadget->normalLabel, app->gc[GC_NORMAL],
		  gadget->x, gadget->y);
  }
}

static void set_state(McGadget *gadget) {
  if (gadget->flags&GAD_SELECTED)
    gadget->flags|=GAD_STATE;
  else
    gadget->flags&=~GAD_STATE;
}

void McGadgetRedraw(McApp *app, McGadget *gadget) {
  set_state(gadget);
  GadgetRedraw(app, gadget, 0);
}

void McGadgetUpdate(McApp *app, McGadget *gadget) {
  set_state(gadget);
  GadgetUpdate(app, gadget, 0);
}

int McGadgetEvent(McApp *app, McGadget *gadget, XEvent *event) {
  int in;

  if (event->type == Expose) {
    int x=gadget->x-app->borderWidth;
    int y=gadget->y-app->borderWidth;
    int x2=x+gadget->width+((app->borderWidth)<<1);
    int y2=y+gadget->height+((app->borderWidth)<<1);
    int ex=event->xexpose.x;
    int ey=event->xexpose.y;
    int ex2=ex+event->xexpose.width;
    int ey2=ey+event->xexpose.height;
    if (ex<x2 && ex2>x && ey<y2 && ey2>y) {
      set_state(gadget);
      GadgetRedraw(app, gadget, 0);
    }
    return 0;
  }

  if (!(gadget->flags&GAD_ACTIVE)) return 0;

  if (gadget->specialInfo)
    return (*(gadget->specialInfo->eventProc))(app, gadget, event);

  /* Standard bool gadget */
  if (gadget->flags&GAD_PRESSED) {
    switch (event->type) {
    case ButtonRelease:
      gadget->flags&=~GAD_PRESSED;
      if (McInRectangle(event->xbutton.x, event->xbutton.y,
			gadget->x, gadget->y, gadget->width, gadget->height)) {
	if (gadget->callbackUp) {
	  if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	    set_state(gadget);
	    GadgetRedraw(app, gadget, 1);
	    XFlush(app->display);
	  }
	  (*gadget->callbackUp)(gadget);
	}
	if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	  if (!(gadget->flags&GAD_TOGGLE)) {
	    gadget->flags^=GAD_SELECTED;
	  }
	}
      } else {
	if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	  gadget->flags^=GAD_SELECTED;
	}
      }
      if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	set_state(gadget);
	GadgetRedraw(app, gadget, 0);
      }
      return 0;
    case MotionNotify:
      in = McInRectangle(event->xmotion.x, event->xmotion.y, gadget->x,
			 gadget->y, gadget->width, gadget->height);
      
      if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	if ((in && (gadget->flags&GAD_SELECTED)) ||
	    ((!in) && (!(gadget->flags&GAD_SELECTED)))) {
	  if (!(gadget->flags&GAD_STATE)) {
	    gadget->flags|=GAD_STATE;
	    GadgetRedraw(app, gadget, 0);
	  }
	} else {
	  if (gadget->flags&GAD_STATE) {
	    gadget->flags&=~GAD_STATE;
	    GadgetRedraw(app, gadget, 0);
	  }
	}
      }
      return 1;
    case ButtonPress:
      return 1;
    default:
      return 0;
    }
  } else { /* not currently pressed */
    switch (event->type) {
    case ButtonPress:
      if (McInRectangle(event->xbutton.x, event->xbutton.y,
			gadget->x, gadget->y, gadget->width, gadget->height)) {
	gadget->flags|=GAD_PRESSED;
	if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	  gadget->flags^=GAD_SELECTED;
	}
	if (gadget->callbackDown) {
	  if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	    set_state(gadget);
	    GadgetRedraw(app, gadget, 1);
	    XFlush(app->display);
	  }
	  (*gadget->callbackDown)(gadget);
	}
	if ((gadget->flags & GAD_HMASK) != GAD_HNONE) {
	  set_state(gadget);
	  GadgetRedraw(app, gadget, 0);
	}
	McMoveGadgetToStart(app, gadget);
    	return 1;
      } else {
	return 0;
      }
    default:
      return 0;
    }
  }
}

void McGadgetInfo(McGadget *gadget) {
  fprintf(stderr,
	  "        x:%d\n"
	  "        y:%d\n"
	  "    width:%d\n"
	  "   height:%d\n"
	  "    flags:$%04x\n"
	  "     type:%d\n",
	  gadget->x, gadget->y, gadget->width, gadget->height,
	  gadget->flags, gadget->type);
}
