/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _McGadget_h_
#define _McGadget_h_

#include "McText.h"
#include "McBitmap.h"

/* Gadget flags */
#define GAD_H3D		0
#define GAD_HCOMP	1
#define GAD_HBOX	2
#define GAD_HNONE	3
#define GAD_HMASK	3
#define GAD_ACTIVE	4	/* This gadget wants events */
#define GAD_SELECTED	8
#define GAD_PRESSED	16
#define GAD_TOGGLE	32
#define GAD_INVIS	64
#define GAD_STATE	128	/* private */
#define GAD_3D		256
#define GAD_DISABLE	512
#define GAD_NOFILL	1024

/* Gadget types */
enum {
  BOOLGADGET,
  SLIDERGADGET,
};

typedef struct McSpecialInfo {
  void	(*updateProc)(struct McApp *, struct McGadget *, int);
  int (*eventProc)(struct McApp *, struct McGadget *, XEvent *);
} McSpecialInfo;

typedef struct McGadget {
  struct McGadget *next;
  struct McGadget *prev;
  unsigned int type;
  unsigned int flags;
  struct McBitmap *normalBitmap;
  struct McBitmap *selectBitmap;
  struct McText *normalLabel;
  struct McText *selectLabel;
  void	(*callbackDown)(struct McGadget *);
  void	(*callbackUp)(struct McGadget *);
  int x, y;
  int width, height;
  int id;
  void *customData;
  struct McSpecialInfo *specialInfo;
} McGadget;

extern McGadget *McCreateGadget(McApp *app, unsigned int flags,
				unsigned int type, int x, int y, int w, int h);
extern void McGadgetRedraw(McApp *app, McGadget *gadget);
extern void McGadgetUpdate(McApp *app, McGadget *gadget);
extern int McGadgetEvent(McApp *app, McGadget *gadget, XEvent *event);
void McGadgetInfo(McGadget *gadget);


#endif /* _McGadget_h_ */

