/* Copyright (C) 1994 
            Olav Woelfelschneider (wosch@rbg.informatik.th-darmstadt.de)

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <X11/Xlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "McApp.h"
#include "McGadget.h"
#include "McSlider.h"

#define SLIDER ((McSlider *)(gadget->specialInfo))

static void McSliderUpdate(McApp *app, McGadget *gadget, int busy);
static int McSliderEvent(McApp *app, McGadget *gadget, XEvent *event);

McSpecialInfo *McCreateSlider(unsigned int flags, unsigned int minValue,
			      unsigned int maxValue) {
  McSlider *slider = (McSlider *) malloc(sizeof(McSlider));
  slider->flags = flags;
  slider->leftValue = slider->rightValue = ((maxValue-minValue)>>1)+minValue;
  slider->minValue = minValue;
  slider->maxValue = maxValue;
  slider->specialInfo.updateProc = McSliderUpdate;
  slider->specialInfo.eventProc = McSliderEvent;
  slider->step = 0;
  return (McSpecialInfo *)slider;
}

static void McSliderUpdate(McApp *app, McGadget *gadget, int busy) {
  int leftPixels;
  int rightPixels;
  int bw;

  bw = app->borderWidth;
  leftPixels=SLIDER->leftValue;
  if (leftPixels>SLIDER->maxValue)
    leftPixels=SLIDER->leftValue = SLIDER->maxValue;
  if (leftPixels<SLIDER->minValue)
    leftPixels=SLIDER->leftValue = SLIDER->minValue;
  rightPixels=SLIDER->rightValue;
  if (rightPixels>SLIDER->maxValue)
    rightPixels=SLIDER->rightValue = SLIDER->maxValue;
  if (rightPixels<SLIDER->minValue)
    rightPixels=SLIDER->rightValue = SLIDER->minValue;

  if (SLIDER->flags & SLIDER_VERTICAL) {
    leftPixels  = leftPixels  * (gadget->height-bw-bw-bw) /
      (SLIDER->maxValue - SLIDER->minValue);
    rightPixels = rightPixels * (gadget->height-bw-bw-bw) /
      (SLIDER->maxValue - SLIDER->minValue);
    if (SLIDER->flags & SLIDER_STEREO) {
      if ((leftPixels-bw)>0) {
	XFillRectangle(app->display, app->window, app->gc[GC_REVERSE],
		       gadget->x+bw+bw,
		       gadget->y+gadget->height-leftPixels-bw,
		       (gadget->width - (bw*6))>>1,
		       leftPixels-bw);
	McAppDrawbox(app, gadget->x+bw+bw,
		     gadget->y+gadget->height-leftPixels-bw,
		     (gadget->width - (bw*6))>>1,
		     leftPixels-bw, _3D_OUT);
      }
      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x+bw,gadget->y+bw,
		     (gadget->width>>1)-bw,
		     gadget->height-leftPixels-bw-bw-bw);

      if ((rightPixels-bw)>0) {
	XFillRectangle(app->display, app->window, app->gc[GC_REVERSE],
		       gadget->x + 3 + gadget->width/2,
		       gadget->y+gadget->height-rightPixels-bw,
		       (gadget->width - (bw*6))>>1,
		       rightPixels-bw);
	McAppDrawbox(app, gadget->x + 3 + gadget->width/2,
		     gadget->y+gadget->height-rightPixels-bw,
		     (gadget->width - (bw*6))>>1,
		     rightPixels-bw, _3D_OUT);
      }
      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x + 1 + gadget->width/2, gadget->y+bw,
		     (gadget->width>>1)-bw,
		     gadget->height-rightPixels-bw-bw-bw);

    } else {
      if ((leftPixels-bw)>0) {
	XFillRectangle(app->display, app->window, app->gc[GC_REVERSE],
		       gadget->x + bw+bw,
		       gadget->y+gadget->height-leftPixels-bw,
		       gadget->width - (4*bw),
		       leftPixels-bw);
	McAppDrawbox(app, gadget->x + bw+bw,
		     gadget->y+gadget->height-leftPixels-bw,
		     gadget->width - (4*bw),
		     leftPixels-bw, _3D_OUT);
      }

      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x+bw,gadget->y+bw,
		     gadget->width-bw-bw,
		     gadget->height-leftPixels-bw-bw-bw);
    }
  } else {
    leftPixels  = leftPixels  * (gadget->width-bw-bw-bw) /
      (SLIDER->maxValue - SLIDER->minValue);
    rightPixels = rightPixels * (gadget->width-bw-bw-bw) /
      (SLIDER->maxValue - SLIDER->minValue);
    if (SLIDER->flags & SLIDER_STEREO) {
      if ((leftPixels-bw)>0) {
	XFillRectangle(app->display, app->window, app->gc[GC_REVERSE],
		       gadget->x+bw+bw, gadget->y+bw+bw,
		       leftPixels - bw,
		       (gadget->height - (bw*6))>>1);
	McAppDrawbox(app, gadget->x+bw+bw, gadget->y+bw+bw,
		     leftPixels - bw,
		   (gadget->height - (bw*6))>>1, _3D_OUT);
      }
      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x+bw+bw+leftPixels, gadget->y+bw,
		     gadget->width-leftPixels-bw-bw-bw,
		     (gadget->height>>1)-bw);
      
      if ((rightPixels-bw)>0) {
	XFillRectangle(app->display, app->window, app->gc[GC_REVERSE],
		       gadget->x+bw+bw, gadget->y + 3 + gadget->height/2,
		       rightPixels - bw,
		       (gadget->height - (bw*6))>>1);
	McAppDrawbox(app, gadget->x+bw+bw, gadget->y + 3 + gadget->height/2,
		     rightPixels - bw,
		     (gadget->height - (bw*6))>>1, _3D_OUT);
      }
      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x+bw+bw+rightPixels,
		     gadget->y+(bw>>1)+gadget->height/2,
		     gadget->width-rightPixels-bw-bw-bw,
		     (gadget->height>>1)-bw);
    } else {
      if ((leftPixels-bw)>0) {
	XFillRectangle(app->display, app->window, app->gc[GC_REVERSE],
		       gadget->x + bw+bw, gadget->y+bw+bw,
		       leftPixels - bw, gadget->height - (4*bw));
	McAppDrawbox(app, gadget->x + bw+bw, gadget->y+bw+bw,
		     leftPixels - bw, gadget->height - (4*bw),
		     _3D_OUT);
      }
      XFillRectangle(app->display, app->window, app->gc[GC_SELECTED],
		     gadget->x+bw+bw+leftPixels, gadget->y+bw,
		     gadget->width-leftPixels-bw-bw-bw,
		     gadget->height-bw-bw);
    }
  }
}

static int CalcValue(McApp *app, McGadget *gadget, int x, int y) {
  int tl, tr, changed;
  int bw;

  bw = app->borderWidth;
  changed = 0;

  if (x>=0) {
    if (SLIDER->flags & SLIDER_VERTICAL) {
      tl = tr = (gadget->y + gadget->height - y) *
	(SLIDER->maxValue-SLIDER->minValue) /
           (gadget->height - (bw*3)) + SLIDER->minValue;
    } else {
      tl = tr = (x - gadget->x - bw) * (SLIDER->maxValue-SLIDER->minValue) /
	(gadget->width - (bw*3)) + SLIDER->minValue;
    }
  } else {
    if (!(tr=SLIDER->step)) tr=y; else tr*=y;
    tl = SLIDER->leftValue + tr;
    tr = SLIDER->rightValue + tr;
  }

  if (tl < SLIDER->minValue) tl = SLIDER->minValue;
  if (tl > SLIDER->maxValue) tl = SLIDER->maxValue;
  if (tr < SLIDER->minValue) tr = SLIDER->minValue;
  if (tr > SLIDER->maxValue) tr = SLIDER->maxValue;

  if (SLIDER->flags & SLIDER_SET_LEFT) {
    if (tl!=SLIDER->leftValue) {
      SLIDER->leftValue=tl;
      changed=1;
    }
  }
  if (SLIDER->flags & SLIDER_SET_RIGHT) {
    if (tr!=SLIDER->rightValue) {
      SLIDER->rightValue=tr;
      changed=1;
    }
  }
  return changed;
}

static int McSliderEvent(McApp *app, McGadget *gadget, XEvent *event) {
  int i,j;

  if (gadget->flags&GAD_PRESSED) {
    switch (event->type) {
    case ButtonRelease:
      gadget->flags&=~GAD_PRESSED;
      SLIDER->flags&=~SLIDER_SET_MASK;
      if (McInRectangle(event->xbutton.x, event->xbutton.y,
			gadget->x+app->borderWidth,
			gadget->y, gadget->width-((app->borderWidth)<<1),
			gadget->height)) {
	if (gadget->callbackUp) {
	  (*gadget->callbackUp)(gadget);
	}
      }
      return 0;

    case MotionNotify:
      if (CalcValue(app, gadget, event->xmotion.x, event->xmotion.y)) {
	McSliderUpdate(app, gadget, 0);
	if (gadget->callbackDown) {
	  XFlush(app->display);
	  (*gadget->callbackDown)(gadget);
	}
      }
      return 1;
    case ButtonPress:
      return 1;
    default:
      return 0;
    }
  } else { /* not currently pressed */
    switch (event->type) {
    case ButtonPress:
      if (McInRectangle(event->xbutton.x, event->xbutton.y,
			gadget->x, gadget->y, gadget->width, gadget->height)) {
	if (SLIDER->flags & SLIDER_STEREO) {
	  SLIDER->flags&=~SLIDER_SET_MASK;
	  if (SLIDER->flags & SLIDER_VERTICAL) {
	    i = (event->xbutton.x - gadget->x);
	    j = gadget->width;
	  } else {
	    i = (event->xbutton.y - gadget->y);
	    j = gadget->height;
	  }
	  if (i < (j/3))
	    SLIDER->flags|=SLIDER_SET_LEFT;
	  else
	    if (i > (j/3*2))
	      SLIDER->flags|=SLIDER_SET_RIGHT;
	    else
	      SLIDER->flags|=SLIDER_SET_LEFT | SLIDER_SET_RIGHT;
	} else {
	  SLIDER->flags|=SLIDER_SET_LEFT | SLIDER_SET_RIGHT;
	}
	switch (event->xbutton.button) {
	case 1:
	  CalcValue(app, gadget, -1, 1);
	  goto btn1_3;
	case 3:
	  CalcValue(app, gadget, -1, -1);
btn1_3:
	  McSliderUpdate(app, gadget, 0);
	  if (gadget->callbackDown) {
	    XFlush(app->display);
	    (*gadget->callbackDown)(gadget);
	  }
	  if (gadget->callbackUp) {
	    XFlush(app->display);
	    (*gadget->callbackUp)(gadget);
	  }
	  gadget->flags&=~GAD_PRESSED;
	  McMoveGadgetToStart(app, gadget);
	  return 0;
	default:
	  CalcValue(app, gadget, event->xbutton.x, event->xbutton.y);
	  McSliderUpdate(app, gadget, 0);
	  if (gadget->callbackDown) {
	    XFlush(app->display);
	    (*gadget->callbackDown)(gadget);
	  }
	  gadget->flags|=GAD_PRESSED;
	  McMoveGadgetToStart(app, gadget);
	  return 1;
	}
      } else {
	return 0;
      }
    default:
      return 0;
    }
  }
}

